from magma.bit import Bit
from magma.bits import Bits
from magma.circuit import coreir_port_mapping
from magma.clock import AsyncReset, AsyncResetN, Clock, _ClockType
from magma.conversions import bit, convertbit
from magma.digital import Digital
from magma.generator import Generator2
from magma.interface import IO
from magma.t import In, Out


def _simulate_wire(self, value_store, state_store):
    value_store.set_value(self.O, value_store.get_value(self.I))


class Wire(Generator2):
    def __init__(self, T):
        if issubclass(T, (AsyncReset, AsyncResetN, Clock)):
            self._gen_named_type_wrapper(T)
            # Standalone return to avoid return-in-init lint warning
            return
        if issubclass(T, Digital):
            coreir_lib = "corebit"
            coreir_genargs = None
        else:
            width = T.flat_length()
            T = Bits[width]
            coreir_lib = "coreir"
            coreir_genargs = {"width": width}
        self.name = "Wire"
        self.io = IO(I=In(T), O=Out(T))
        self.simulate = _simulate_wire
        self.coreir_name = "wire"
        self.coreir_lib = coreir_lib
        self.coreir_genargs = coreir_genargs
        self.renamed_ports = coreir_port_mapping

    def _gen_named_type_wrapper(self, T):
        """
        Generates a container around Wire(Bit) so named types are wrapped
        properly
        """
        assert issubclass(T, Digital)
        self.io = IO(I=In(T), O=Out(T))
        self.name = f"Wire{T}"
        self.io.O @= convertbit(Wire(Bit)()(bit(self.io.I)), T)



from typing import Union
from magma.bit import Bit
from magma.logging import root_logger
from magma.primitives.mux import mux
from magma.t import Type, In, Out
from magma.protocol_type import MagmaProtocol
from magma.tuple import Product, ProductKind


_logger = root_logger()


class ReadyValidException(Exception):
    pass


class ReadyValidKind(ProductKind):
    def _check_T(cls, T):
        if not issubclass(T, (Type, MagmaProtocol)):
            raise TypeError(
                f"{cls}[T] expected T to be a subclass of m.Type or"
                f" m.MagmaProtocol not {T}"
            )
        undirected_T = T.undirected_t
        if undirected_T is not T:
            _logger.warning(
                f"Type {T} used with ReadyValid is not undirected, converting"
                " to undirected type"
            )
        return undirected_T

    def __getitem__(cls, T: Union[Type, MagmaProtocol]):
        T = cls._check_T(T)
        fields = {"valid": Bit, "data": T, "ready": Bit}
        return type(f"{cls}[{T}]", (cls, ), fields)

    def flip(cls):
        raise TypeError("Cannot flip an undirected ReadyValid type")


class ReadyValid(Product, metaclass=ReadyValidKind):
    """
    An I/O Bundle containing 'valid' and 'ready' signals that handshake
    the transfer of data stored in the 'data' subfield.

    The base protocol is undirected and is meant to be used with the modifiers:
    * m.Consumer(m.ReadyValid[T])  # data/valid are inputs, ready is output
    * m.Producer(m.ReadyValid[T])  # data/valid are outputs, ready is input

    T is the type of data to be wrapped in Ready/Valid

    The actual semantics of ready/valid are enforced via the use of concrete
    subclasses.
    """
    def fired(self):
        """
        Returns a bit that is high when ready and valid are both high
        """
        raise NotImplementedError


class ReadyValidProducerKind(ReadyValidKind):
    def __getitem__(cls, T: Union[Type, MagmaProtocol]):
        T = cls._check_T(T)
        fields = {"valid": Out(Bit), "data": Out(T), "ready": In(Bit)}
        return type(f"{cls}[{T}]", (cls, ), fields)

    def flip(cls):
        return Consumer(cls)

    def __str__(cls):
        if not cls.is_bound:
            return cls.__name__
        return f"Producer(ReadyValid[{cls.data.undirected_t()}])"


class ReadyValidProducer(ReadyValid, metaclass=ReadyValidProducerKind):
    def fired(self):
        """
        Returns a bit that is high when ready and valid are both high
        """
        if not self.ready.driven():
            raise ReadyValidException(
                "Cannot invoke fired on producer when ready isn't driven"
            )
        return self.ready.value() & self.valid

    def deq(self, when=True):
        """
        Assert ready on this port and return associated data when `when` is
        high

        This is typically used when valid has been asserted on the producer
        side.
        """
        if when is not True:
            if not self.ready.driven():
                raise ReadyValidException(
                    "Cannot use deq with when without a default ready driver"
                )
            ready = self.ready.value()
            self.ready.unwire(ready)
            self.ready @= ready | when
        else:
            self.ready @= when
        return self.data

    def no_deq(self, when=True):
        """
        Indicates no dequeue occurs, ready is set to False
        """
        if when is not True:
            if not self.ready.driven():
                raise ReadyValidException(
                    "Cannot use no_deq with condition without an existing"
                    " ready driver"
                )
            ready = self.ready.value()
            self.ready.unwire(ready)
            self.ready @= mux([ready, False], when)
        else:
            self.ready @= False


class ReadyValidConsumerKind(ReadyValidKind):
    def __getitem__(cls, T: Union[Type, MagmaProtocol]):
        T = cls._check_T(T)
        fields = {"valid": In(Bit), "data": In(T), "ready": Out(Bit)}
        return type(f"{cls}[{T}]", (cls, ), fields)

    def flip(cls):
        return Producer(cls)

    def __str__(cls):
        if not cls.is_bound:
            return cls.__name__
        return f"Consumer(ReadyValid[{cls.data.undirected_t()}])"


class ReadyValidConsumer(ReadyValid, metaclass=ReadyValidConsumerKind):
    def fired(self):
        if not self.valid.driven():
            raise ReadyValidException(
                "Cannot invoke fired on consumer when valid isn't driven"
            )

        return self.ready & self.valid.value()

    def enq(self, value, when=True):
        """
        Produces the data `value` when `when` is high

        Uses "last connect semantics", which allows multiple `enq` methods to
        be invoked with different conditions with precedence given to the later
        invocations.
        """
        if when is not True:
            if not (self.valid.driven() and self.data.driven()):
                raise ReadyValidException(
                    "Cannot use enq when without default driver"
                )
            curr_valid = self.valid.value()
            when |= curr_valid
            self.valid.unwire(curr_valid)

            curr_data = self.data.value()
            value = mux([curr_data, value], when)
            self.data.unwire(curr_data)
        self.valid @= when
        self.data @= value
        return value

    def no_enq(self, when=True):
        """
        Indicates no enqueue occurs.  Valid is set to false, and bits are
        connected to 0 (TODO: Should be an unitialized wire (Don't Care))
        """
        if when is not True:
            if not (self.valid.driven() and self.data.driven()):
                raise ReadyValidException(
                    "Cannot use no_deq with condition without an existing"
                    "valid and data driver"
                )
            valid = self.valid.value()
            data = self.data.value()
            self.valid.unwire(valid)
            self.data.unwire(data)
            self.valid @= mux([valid, False], when)
            self.data @= mux([data, 0], when)
        else:
            self.valid @= False
            self.data @= 0


class Decoupled(ReadyValid):
    """
    `valid` indicates the prdoucer has put valid data in `data`, and `ready`
    indicates that the consumer is ready to accept the data this cycle.  No
    requirements are made on the signaling of ready or valid.
    """
    # TODO: Add down convert logic from Irrevocable consumer to Decoupled,
    # dropping guarantees of irrevocability.  Cannot be used on producer
    pass


class DecoupledConsumer(ReadyValidConsumer, Decoupled):
    pass


class DecoupledProducer(ReadyValidProducer, Decoupled):
    pass


class EnqIO(DecoupledProducer):
    """
    Alias for Producer(Decoupled[T])

    drives (outputs) `valid` and `data`, inputs `ready`
    """
    pass


class DeqIO(DecoupledConsumer):
    """
    Alias for Consumer(Decoupled[T])

    drives (outputs) `ready`, inputs `valid` and `data`
    """
    pass


class Irrevocable(ReadyValid):
    """
    Promises not to change the value of `data` after a cycle where `valid` is
    high and `ready` is low.  Additionally, once `valid` is raised, it will
    never be lowered until after `ready` has also been raised
    """
    # TODO: Add upconvert logic from Decoupled producer to Irrevocable (can use
    # Irrevocable where a Decoupled is expected). Cannot be used on consumer.
    pass


class IrrevocableConsumer(ReadyValidConsumer, Irrevocable):
    pass


class IrrevocableProducer(ReadyValidProducer, Irrevocable):
    pass


def Consumer(T: ReadyValidKind):
    if issubclass(T, ReadyValid):
        undirected_T = T.data.undirected_t
    if issubclass(T, Irrevocable):
        return IrrevocableConsumer[undirected_T]
    if issubclass(T, Decoupled):
        return DecoupledConsumer[undirected_T]
    if issubclass(T, ReadyValid):
        return ReadyValidConsumer[undirected_T]
    raise TypeError(f"Consumer({T}) is unsupported")


def Producer(T: ReadyValidKind):
    if issubclass(T, ReadyValid):
        undirected_T = T.data.undirected_t
    if issubclass(T, Irrevocable):
        return IrrevocableProducer[undirected_T]
    if issubclass(T, Decoupled):
        return DecoupledProducer[undirected_T]
    if issubclass(T, ReadyValid):
        return ReadyValidProducer[undirected_T]
    raise TypeError(f"Consumer({T}) is unsupported")


# TODO: QueueIO and Queue



from abc import ABC, abstractmethod
from enum import Enum, auto
from ..bit import Bit
from ..bits import Bits
from ..circuit import DeclareCircuit, DefineCircuit
from ..digital import Digital
from ..logging import root_logger


_logger = root_logger()


class ImportMode(Enum):
    DECLARE = auto()
    DEFINE = auto()

    def func(self):
        if self is ImportMode.DECLARE:
            return DeclareCircuit
        return DefineCircuit


class VerilogImportError(Exception):
    pass


class MultiplePortDeclarationError(VerilogImportError):
    def __init__(self, name):
        super().__init__(name)

    def __str__(self):
        name = self.args[0]
        return f"Multiple declarations of port '{name}'"


class MultipleModuleDeclarationError(VerilogImportError):
    def __init__(self, name):
        super().__init__(name)

    def __str__(self):
        name = self.args[0]
        return f"Multiple definitions of module '{name}'"


class MixedPortDeclarationsError(VerilogImportError):
    def __init__(self):
        super().__init__()

    def __str__(self):
        return "Can not parse mixed declared and undeclared types"


class VerilogImporter(ABC):
    def __init__(self, type_map):
        self._type_map = type_map
        self._magma_modules = []

    def reset(self, type_map=None, external_modules=None):
        if type_map is not None:
            self._type_map = type_map
        self._magma_modules = []

    def map_type(self, name, magma_type):
        if name not in self._type_map:
            return magma_type
        target_type = self._type_map[name]
        if issubclass(magma_type, Digital) and \
           issubclass(target_type, Digital) and \
           magma_type.direction == target_type.direction:
            return target_type
        if issubclass(magma_type, Bits) and \
           issubclass(target_type, Bits) and \
           magma_type.N == target_type.N and \
           magma_type.T.direction == target_type.T.direction:
            return target_type
        raise NotImplementedError(f"Conversion between {magma_type} and "
                                  f"{target_type} not supported")

    def add_module(self, module):
        self._magma_modules.append(module)

    @abstractmethod
    def import_(self, src, mode):
        raise NotImplementedError()



import hashlib
import string
from ast_tools.stack import _SKIP_FRAME_DEBUG_STMT, get_symbol_table
from magma.circuit import _definition_context_stack, Circuit, IO
from magma.passes.passes import CircuitPass
from magma.primitives.wire import Wire
from magma.t import Type, Direction, In
from magma.view import PortView, InstView
from magma.digital import Digital
from magma.bit import Bit
from magma.array import Array
from magma.tuple import Tuple
from magma.wire import wire
from magma.ref import DefnRef, InstRef, ArrayRef, TupleRef


def _get_top_level_ref(ref):
    if isinstance(ref, ArrayRef):
        return _get_top_level_ref(ref.array.name)
    if isinstance(ref, TupleRef):
        return _get_top_level_ref(ref.tuple.name)
    return ref


def _get_view_inst_parent(view):
    while not isinstance(view, InstView):
        assert isinstance(view, PortView), type(view)
        return _get_view_inst_parent(view.parent)
    return view


def _make_inline_value(cls, inline_value_map, value):
    if isinstance(value, Array) and not issubclass(value.T, Digital):
        return "'{" + ", ".join(_make_inline_value(cls, inline_value_map, t)
                                for t in reversed(value)) + "}"
    if isinstance(value, Tuple):
        raise NotImplementedError("Inlining tuple to verilog")
    value_key = f"__magma_inline_value_{len(inline_value_map)}"
    inline_value_map[value_key] = value
    return f"{{{value_key}}}"


def _make_temporary(defn, value, num, inline_wire_prefix, parent=None):
    with defn.open():
        # Insert a wire so it can't be inlined out
        temp_name = f"{inline_wire_prefix}{num}"
        temp = Wire(type(value).undirected_t)(name=temp_name)
        temp.I @= value
    if parent is not None:
        return PortView(temp.O, parent)
    return temp.O


def _insert_temporary_wires(cls, value, inline_wire_prefix):
    """
    For non DefnRef, insert a temporary Wire instance so the signal isn't
    inlined out
    """
    if isinstance(value, Type):
        if value.is_input():
            value = value.value()
        if not isinstance(_get_top_level_ref(value.name), DefnRef):
            if value not in cls.inline_verilog_wire_map:
                temp = _make_temporary(cls, value,
                                       len(cls.inline_verilog_wire_map),
                                       inline_wire_prefix)
                cls.inline_verilog_wire_map[value] = temp
            value = cls.inline_verilog_wire_map[value]
    else:
        assert isinstance(value, PortView)
        if value.port.is_input():
            # For now we can't handle input port views, since this might return
            # a connection to the interface (e.g. self.I), then we'd need to
            # walk up to the parent instance and select the corresponding input
            # port, then trace that, which may need to happen recursively until
            # we find the driver in the current definition.
            # We don't have any existing use case for this, and a work around
            # exists which requires the user to effectively use the driver we
            # would find, rather than selecting the hierarchical input port
            # (this should work for simple cases, but might be annoying for
            # deep hierarchies where the driver could come from a far away
            # place, at which point we can add this feature)
            raise NotImplementedError()
        if value not in cls.inline_verilog_wire_map:
            # get first instance parent, then the parent of that will
            # be the container where we insert a wire
            parent = _get_view_inst_parent(value).parent
            if isinstance(parent, InstView):
                defn = parent.circuit
            else:
                assert isinstance(parent, Circuit)
                defn = type(parent)

            temp = _make_temporary(defn, value.port,
                                   len(cls.inline_verilog_wire_map),
                                   inline_wire_prefix, parent)
            cls.inline_verilog_wire_map[value] = temp
        value = cls.inline_verilog_wire_map[value]
    return value


def _build_io(inline_value_map):
    if not inline_value_map:
        # if there's not inlined values, make a dummy input/instance so there's
        # a def
        io = IO(I=In(Bit))
        io.I.unused()
        return io
    io = IO()
    for key, value in inline_value_map.items():
        if isinstance(value, PortView):
            T = value.T
        else:
            T = type(value)
        io += IO(**{key: In(T)})
    return io


def _inline_verilog(cls, inline_str, inline_value_map, inline_wire_prefix,
                    **kwargs):
    format_args = {}
    for key, arg in kwargs.items():
        if isinstance(arg, (Type, PortView)):
            arg = _make_inline_value(cls, inline_value_map, arg)
        format_args[key] = arg
    inline_str = inline_str.format(**format_args)

    # modules/instances will be sorted lexigraphically in the generated
    # verilog, so we need to generate a lexicographically increasing string
    # based on the module count
    i = len(cls.inline_verilog_modules)

    # For every 10 modules, we insert a prefix 9 (so it comes after the
    # previous 1-9 digits)
    prefix_len = i // 10
    prefix = "9" * prefix_len

    suffix = i % 10

    class _InlineVerilog(Circuit):

        # Unique name (hash) since uniquify doesn't check inline_verilog
        name = f"{cls.name}_inline_verilog_{prefix}{suffix}"

        io = _build_io(inline_value_map)

        # each inline value (key) will map to a port, populate the
        # connect_references dictionary with a map from key to port
        connect_references = {}

        for key in inline_value_map:
            port = getattr(io, key)
            # Need to mark unused so coreir knows there's a module definition
            port.unused()
            connect_references[key] = port

        inline_verilog_strs = [(inline_str, connect_references)]

    cls.inline_verilog_modules.append(_InlineVerilog)

    with cls.open():
        inst = _InlineVerilog(
            name=f"{cls.name}_inline_verilog_inst_{prefix}{suffix}"
        )
        # Dummy var so there's a defn for inline verilog without any
        # interpoalted avlues
        if not inline_value_map:
            inst.I @= 0

    for key, value in inline_value_map.items():
        value = _insert_temporary_wires(cls, value, inline_wire_prefix)
        wire(value, getattr(inst, key))


def _process_fstring_syntax(format_str, format_args, cls, inline_value_map,
                            symbol_table):
    fieldnames = [fname
                  for _, fname, _, _ in string.Formatter().parse(format_str)
                  if fname]
    for field in fieldnames:
        if field in format_args:
            continue
        try:
            value = eval(field, {}, symbol_table)
        except NameError:
            continue
        if isinstance(value, (Type, PortView)):
            # These have special handling, don't convert to string.
            value = _make_inline_value(cls, inline_value_map, value)
            # Stage for subsequent format call for regular kwargs inside
            # _inline_verilog
            value = value.replace("{", "{{").replace("}", "}}")
        format_str = format_str.replace(f"{{{field}}}", str(value))
    return format_str


def _process_inline_verilog(cls, format_str, format_args, symbol_table,
                            inline_wire_prefix):
    inline_value_map = {}
    if symbol_table is not None:
        format_str = _process_fstring_syntax(format_str, format_args, cls,
                                             inline_value_map, symbol_table)
    _inline_verilog(cls, format_str, inline_value_map, inline_wire_prefix,
                    **format_args)


class ProcessInlineVerilogPass(CircuitPass):
    def __call__(self, cls):
        if cls.inline_verilog_generated:
            return
        if not cls._context_._inline_verilog:
            return
        cls.inline_verilog_wire_map = {}
        cls.inline_verilog_modules = []
        with cls.open():
            for fields in cls._context_._inline_verilog:
                _process_inline_verilog(cls, *fields)
        if getattr(cls, "instances", []):
            cls._is_definition = True
        cls.inline_verilog_generated = True


def inline_verilog(format_str, inline_wire_prefix="_magma_inline_wire",
                   **kwargs):
    exec(_SKIP_FRAME_DEBUG_STMT)
    symbol_table = get_symbol_table([inline_verilog], copy_locals=True)

    context = _definition_context_stack.peek()
    context.add_inline_verilog(format_str, kwargs, symbol_table,
                               inline_wire_prefix)



import functools


def output_only(msg):

    def _wrapper(fn):

        @functools.wraps(fn)
        def _wrapped(self, *args, **kwargs):
            if self.is_input():
                raise TypeError(msg)
            return fn(self, *args, **kwargs)

        return _wrapped

    return _wrapper



import enum
import functools
from typing import Optional

from magma.circuit import CircuitKind, CircuitType, DefineCircuitKind
from magma.conversions import as_bits
from magma.value_utils import ValueVisitor


class StubType(enum.Enum):
    ZERO = enum.auto()


class _StubifyVisitor(ValueVisitor):
    def __init__(self, stub_type: StubType):
        self._stub_type = stub_type
        self._modified = False

    @property
    def modified(self):
        return self._modified

    def _drive_output(self, value):
        if self._stub_type == StubType.ZERO:
            bits = as_bits(value)
            bits @= 0
            return
        raise NotImplementedError(self._stub_type)

    def _handle_value(self, value, leaf=False):
        if value.is_output():
            self._modified = True
            value.unused()
            return
        if value.is_input():
            self._modified = True
            self._drive_output(value)
            return
        if leaf:
            return
        super().generic_visit(value)

    def generic_visit(self, value):
        self._handle_value(value, leaf=False)

    def visit_Bits(self, value):
        self._handle_value(value, leaf=True)

    def visit_Digital(self, value):
        self._handle_value(value, leaf=True)


def _stubify_impl(ckt: CircuitKind, stub_type: StubType):
    modified = False
    for port in ckt.interface.ports.values():
        visitor = _StubifyVisitor(stub_type)
        visitor.visit(port)
        modified |= visitor.modified
    if modified:
        # NOTE(rsetaluri): This is a hack because we don't have good handling of
        # isdefinition when the circuit is modified. We should be doing that
        # more principled-ly. See https://github.com/phanrahan/magma/issues/929.
        ckt._is_definition = True


def _stub_open(cls):
    raise NotImplementedError("Can not call open() on a circuit stub")


def stubify(ckt: CircuitKind, stub_type: StubType):
    with ckt.open():
        _stubify_impl(ckt, stub_type)
    # Set ckt.open to be a function which raises a NotImplementedError. Note
    # that we *can't* do this in the class itself, since we need to call open()
    # to tie the outputs first (in stubify()). Afterwards, we can override the
    # method.
    setattr(ckt, "open", classmethod(_stub_open))


def circuit_stub(cls=None, *, stub_type: Optional[StubType] = StubType.ZERO):
    """
    Inspired by https://pybit.es/decorator-optional-argument.html.

    Stub-ifys a circuit, based on optional parameter @stub_type. Default
    behavior is driving all outputs to zero. All inputs will be marked as unused
    automatically as well.
    """
    if cls is None:
        return partial(circuit_stub, stub_type=stub_type)
    stubify(cls, stub_type)
    return cls


class _CircuitStubMeta(DefineCircuitKind):
    def __new__(metacls, name, bases, dct):
        cls = super().__new__(metacls, name, bases, dct)
        # Only call stubify() on user circuits (i.e. do not call on CircuitStub
        # base class).
        if not dct.get("_circuit_base_", False):
            stubify(cls, stub_type=StubType.ZERO)
        return cls


class CircuitStub(CircuitType, metaclass=_CircuitStubMeta):
    _circuit_base_ = True



import abc
import copy
import enum
import inspect
import operator

from magma.bits import Bits, BitsMeta, SInt, reduce
from magma.conversions import uint, bits, sint
from magma.conversions import concat as bits_concat
from magma.debug import debug_wire
from magma.protocol_type import MagmaProtocolMeta, MagmaProtocol
from magma.t import Direction
from magma.type_utils import TypeTransformer, isuint, issint
from magma.value_utils import ValueVisitor, make_selector


def _is_int(value):
    try:
        int(value)
    except (ValueError, TypeError):
        return False
    return True


class Context:
    def __init__(self, assignee, root):
        self.assignee = assignee
        self.root = root

    def max_width(self):

        def _visit(node):
            if hasattr(node, "_width_"):
                return node._width_
            if isinstance(node, _SmartBitsExpr):
                return len(node.bits)
            return max(_visit(arg) for arg in node.args)

        max_width = _visit(self.root)
        if self.assignee is not None:
            max_width = max(max_width, len(self.assignee.bits))
        return max_width

    def __str__(self):
        return f"Context(lhs={repr(self.assignee)}, rhs={repr(self.root)})"


class _SmartExprMeta(MagmaProtocolMeta):
    pass


class _SmartExpr(MagmaProtocol, metaclass=_SmartExprMeta):
    @abc.abstractmethod
    def resolve(self, context):
        raise NotImplementedError()

    @abc.abstractmethod
    def eval(self):
        raise NotImplementedError()

    # Binary arithmetic operators.
    def __add__(self, other: '_SmartExpr'):
        if not isinstance(other, _SmartExpr):
            return NotImplemented
        return _SmartBinaryOpExpr(operator.add, self, other)

    def __sub__(self, other: '_SmartExpr'):
        if not isinstance(other, _SmartExpr):
            return NotImplemented
        return _SmartBinaryOpExpr(operator.sub, self, other)

    def __mul__(self, other: '_SmartExpr'):
        if not isinstance(other, _SmartExpr):
            return NotImplemented
        return _SmartBinaryOpExpr(operator.mul, self, other)

    def __floordiv__(self, other: '_SmartExpr'):
        if not isinstance(other, _SmartExpr):
            return NotImplemented
        return _SmartBinaryOpExpr(operator.floordiv, self, other)

    def __truediv__(self, other: '_SmartExpr'):
        if not isinstance(other, _SmartExpr):
            return NotImplemented
        return _SmartBinaryOpExpr(operator.truediv, self, other)

    def __mod__(self, other: '_SmartExpr'):
        if not isinstance(other, _SmartExpr):
            return NotImplemented
        return _SmartBinaryOpExpr(operator.mod, self, other)

    # Binary logic operators.
    def __and__(self, other: '_SmartExpr'):
        if not isinstance(other, _SmartExpr):
            return NotImplemented
        return _SmartBinaryOpExpr(operator.and_, self, other)

    def __or__(self, other: '_SmartExpr'):
        if not isinstance(other, _SmartExpr):
            return NotImplemented
        return _SmartBinaryOpExpr(operator.or_, self, other)

    def __xor__(self, other: '_SmartExpr'):
        if not isinstance(other, _SmartExpr):
            return NotImplemented
        return _SmartBinaryOpExpr(operator.xor_, self, other)

    # Comparison operators.
    def __eq__(self, other: '_SmartExpr'):
        if not isinstance(other, _SmartExpr):
            return NotImplemented
        return _SmartComparisonOpExpr(operator.eq, self, other)

    def __ne__(self, other: '_SmartExpr'):
        if not isinstance(other, _SmartExpr):
            return NotImplemented
        return _SmartComparisonOpExpr(operator.ne, self, other)

    def __ge__(self, other: '_SmartExpr'):
        if not isinstance(other, _SmartExpr):
            return NotImplemented
        return _SmartComparisonOpExpr(operator.ge, self, other)

    def __gt__(self, other: '_SmartExpr'):
        if not isinstance(other, _SmartExpr):
            return NotImplemented
        return _SmartComparisonOpExpr(operator.gt, self, other)

    def __le__(self, other: '_SmartExpr'):
        if not isinstance(other, _SmartExpr):
            return NotImplemented
        return _SmartComparisonOpExpr(operator.le, self, other)

    def __lt__(self, other: '_SmartExpr'):
        if not isinstance(other, _SmartExpr):
            return NotImplemented
        return _SmartComparisonOpExpr(operator.lt, self, other)

    # Shift operators.
    def __lshift__(self, other: '_SmartExpr'):
        if not isinstance(other, _SmartExpr):
            return NotImplemented
        return _SmartShiftOpExpr(operator.lshift, self, other)

    def __rshift__(self, other: '_SmartExpr'):
        if not isinstance(other, _SmartExpr):
            return NotImplemented
        return _SmartShiftOpExpr(operator.rshift, self, other)

    # Unary operators.
    def __invert__(self):
        return _SmartUnaryOpExpr(operator.invert, self)

    def __neg__(self):
        return _SmartUnaryOpExpr(operator.neg, self)

    # Reduction operators.
    def reduce(self, op):
        return _SmartReductionOpExpr(op, self)

    # Extension operators.
    def zext(self, width):
        return _SmartExtendOpExpr(width, False, self, resolved=False)

    def sext(self, width):
        return _SmartExtendOpExpr(width, True, self, resolved=False)


class _SmartOpExpr(_SmartExpr, metaclass=_SmartExprMeta):
    def __init__(self, op, *args):
        self._op = op
        self._args = args

    @property
    def args(self):
        return self._args

    @property
    def op(self):
        return self._op

    def _update(self, *args):
        self._args = args

    def _resolve_args(self, context):
        for arg in self._args:
            arg.resolve(context)

    def _eval_args(self):
        return [arg.eval() for arg in self._args]

    def __str__(self):
        if inspect.isbuiltin(self._op):
            op_str = self._op.__name__
        else:
            op_str = str(self._op)
        args = ", ".join(str(arg) for arg in self._args)
        return f"{op_str}({args})"


class _SmartExtendOpExpr(_SmartOpExpr):

    class _ExtendOp:
        def __init__(self, width, signed):
            self._width = width
            self._signed = signed

        def __call__(self, operand):
            if self._signed:
                return sint(operand).sext(self._width)
            return uint(operand).zext(self._width)

        def __str__(self):
            return f"Extend[width={self._width}, signed={self._signed}]"

    def __init__(self, width, signed, operand, resolved=True):
        extend = _SmartExtendOpExpr._ExtendOp(width, signed)
        super().__init__(extend, operand)
        self._resolved = resolved

    def resolve(self, context):
        if self._resolved:
            return
        context = Context(None, self)
        self._resolve_args(context)
        self._width_ = self._args[0]._width_ + self.op._width
        self._signed_ = self.op._signed

    def eval(self):
        args = self._eval_args()
        return self.op(*args)


def _extend_if_needed(expr, to_width, signed):
    diff = expr._width_ - to_width
    assert diff <= 0
    if diff == 0:
        return expr
    expr = _SmartExtendOpExpr(-diff, signed, expr)
    expr._width_ = to_width
    expr._signed_ = signed
    return expr


class _SmartReductionOpExpr(_SmartOpExpr):

    class _ReductionOp:
        _OP_TO_NAME = {
            operator.and_: "And",
            operator.or_: "Or",
            operator.xor: "Xor",
        }

        def __init__(self, op):
            cls = _SmartReductionOpExpr._ReductionOp
            if op not in cls._OP_TO_NAME:
                raise ValueError(f"Reduction operator {op} not supported")
            self._op = op

        def __call__(self, operand):
            return reduce(self._op, operand)

        def __str__(self):
            cls = _SmartReductionOpExpr._ReductionOp
            op_name = cls._OP_TO_NAME[self._op]
            return f"{op_name}Reduce"

    def __init__(self, op, operand):
        reduction = _SmartReductionOpExpr._ReductionOp(op)
        super().__init__(reduction, operand)

    def resolve(self, context):
        context = Context(None, self)
        self._resolve_args(context)
        self._width_ = 1
        self._signed_ = all(arg._signed_ for arg in self._args)

    def eval(self):
        args = self._eval_args()
        fn = sint if self._signed_ else uint
        return fn(self.op(*args))


class _SmartNAryContextualOpExr(_SmartOpExpr):
    def __init__(self, op, *args):
        super().__init__(op, *args)

    def resolve(self, context):
        self._resolve_args(context)
        self._signed_ = all(arg._signed_ for arg in self._args)
        to_width = context.max_width()
        args = (_extend_if_needed(arg, to_width, self._signed_)
                for arg in self._args)
        self._update(*args)
        self._width_ = to_width

    def eval(self):
        args = self._eval_args()
        if self._signed_:
            args = (sint(arg) for arg in args)
        else:
            args = (uint(arg) for arg in args)
        return self.op(*args)


class _SmartBinaryOpExpr(_SmartNAryContextualOpExr):
    def __init__(self, op, loperand, roperand):
        super().__init__(op, loperand, roperand)


class _SmartUnaryOpExpr(_SmartNAryContextualOpExr):
    def __init__(self, op, operand):
        super().__init__(op, operand)


class _SmartComparisonOpExpr(_SmartOpExpr):
    def __init__(self, op, loperand, roperand):
        super().__init__(op, loperand, roperand)

    def resolve(self, context):
        context = Context(None, self)
        self._resolve_args(context)
        signed_args = all(arg._signed_ for arg in self._args)
        to_width = max(arg._width_ for arg in self._args)
        args = (_extend_if_needed(arg, to_width, signed_args)
                for arg in self._args)
        self._update(*args)
        self._width_ = 1
        self._signed_ = False

    def eval(self):
        args = self._eval_args()
        signed_args = all(arg._signed_ for arg in self._args)
        fn = sint if signed_args else uint
        args = (fn(arg) for arg in args)
        return uint(self.op(*args))


class _SmartShiftOpExpr(_SmartOpExpr):
    def __init__(self, op, loperand, roperand):
        super().__init__(op, loperand, roperand)

    def resolve(self, context):
        loperand, roperand = self._args
        loperand.resolve(context)
        self_context = Context(None, self)
        roperand.resolve(self_context)
        to_width = max(arg._width_ for arg in self._args)
        self._signed_ = all(arg._signed_ for arg in self._args)
        args = (_extend_if_needed(arg, to_width, self._signed_)
                for arg in self._args)
        self._update(*args)
        self._width_ = to_width

    def eval(self):
        args = self._eval_args()
        if self._signed_:
            args = (sint(arg) for arg in args)
        else:
            args = (uint(arg) for arg in args)
        return self.op(*args)


class _SmartConcatOpExpr(_SmartOpExpr):

    class _ConcatOp:
        def __call__(self, *args):
            return bits_concat(*args)

        def __str__(self):
            return "Concat"

    def __init__(self, *args):
        concat = _SmartConcatOpExpr._ConcatOp()
        super().__init__(concat, *args)

    def resolve(self, context):
        for arg in self._args:
            context = Context(None, arg)
            arg.resolve(context)
        self._width_ = sum(arg._width_ for arg in self._args)
        self._signed_ = False

    def eval(self):
        args = self._eval_args()
        args = [uint(arg) for arg in args]
        return uint(self.op(*args))


def concat(*args):
    if not all(isinstance(arg, _SmartExpr) for arg in args):
        types = ", ".join(str(type(arg)) for arg in args)
        raise NotImplementedError(f"Concat not supported for {types}")
    return _SmartConcatOpExpr(*args)


class _SmartSignedOpExpr(_SmartOpExpr):

    class _SignedOp:
        def __init__(self, signed):
            self._signed = signed

        @property
        def signed(self):
            return self._signed

        def __str__(self):
            return "Signed" if self._signed else "Unsigned"

    def __init__(self, signed, operand):
        signed = _SmartSignedOpExpr._SignedOp(signed)
        super().__init__(signed, operand)

    def resolve(self, context):
        self._resolve_args(context)
        self._width_ = self._args[0]._width_
        self._signed_ = self._op.signed

    def eval(self):
        args = self._eval_args()
        fn = sint if self._signed_ else uint
        return fn(args[0])


def signed(expr):
    return _SmartSignedOpExpr(True, expr)


def unsigned(expr):
    return _SmartSignedOpExpr(False, expr)


class _SmartBitsExpr(_SmartExpr, metaclass=_SmartExprMeta):
    def __init__(self, bits):
        self._bits = bits

    @property
    def bits(self):
        return self._bits

    def __str__(self):
        return str(self._bits)

    def resolve(self, context):
        self._width_ = len(self._bits)
        self._signed_ = type(self._bits)._signed

    def eval(self):
        return self._bits.typed_value()


class _SmartBitsMeta(_SmartExprMeta):
    def __getitem__(cls, key):
        if issubclass(cls, SmartBits) and (hasattr(cls, "_T") or
                                           hasattr(cls, "_signed")):
            raise TypeError("Can not doubly qualify SmartBits, i.e. "
                            "SmartBits[n][m] not allowed")
        signed = False
        if isinstance(key, tuple):
            if len(key) == 1:
                pass
            elif len(key) == 2:
                key, signed = key
            else:
                raise ValueError(f"{key} unsupported")
        if _is_int(key):
            T = Bits[key]
        else:
            assert isinstance(key, BitsMeta)
            T = key
        name = f"SmartBits[{len(T)}, {signed}]"
        dct = {"_T": T, "_signed": signed}
        return type(cls)(name, (cls,), dct)

    def _to_magma_(cls):
        return cls._T

    def _qualify_magma_(cls, d):
        return SmartBits[cls._T.qualify(d), cls._signed]

    def _flip_magma_(cls):
        return SmartBits[cls._T.flip(), cls._signed]

    def _from_magma_value_(cls, value):
        return cls(value)

    def __repr__(cls):
        has_T = hasattr(cls, "_T")
        has_signed = hasattr(cls, "_signed")
        if not (has_T or has_signed):
            return "SmartBits"
        assert has_T and has_signed
        return f"SmartBits[{len(cls._T)}, {cls._signed}]"


class SmartBits(_SmartBitsExpr, metaclass=_SmartBitsMeta):
    def __init__(self, value=None):
        super().__init__(self)
        if value is None:
            value = type(self)._to_magma_()()
        self._value = bits(value)

    def typed_value(self):
        if type(self)._signed:
            return sint(self._value)
        return uint(self._value)

    def untyped_value(self):
        return self._value

    def __hash__(self):
        return hash(self._value)

    def _get_magma_value_(self):
        return self.untyped_value()

    def __deepcopy__(self, memo):
        return type(self)(self._value)

    @debug_wire
    def wire(self, other, debug_info):
        if isinstance(other, Bits):
            super().wire(other, debug_info)
            return
        if not isinstance(other, _SmartExpr):
            raise ValueError(f"Can not wire {type(self)} to {type(other)}")
        evaluated, resolved = _eval(self, other)
        evaluated = evaluated.force_width(len(self))
        MagmaProtocol.wire(self, evaluated)
        self._smart_expr_ = resolved  # attach debug info

    def __len__(self):
        return len(type(self)._T)

    def force_width(self, width):
        diff = len(self) - width
        if diff == 0:
            return copy.deepcopy(self)
        value = self.typed_value()
        value = value.ext(-diff) if diff < 0 else value[:-diff]
        return SmartBits.from_bits(value)

    @staticmethod
    def from_bits(value):
        assert isinstance(value, Bits)
        signed = isinstance(value, SInt)
        return SmartBits[len(value), signed](value)

    def __str__(self):
        signed = type(self)._signed
        return f"SmartBits[{len(self)}, {signed}]({str(self._value)})"


SmartBit = SmartBits[1]


def _eval(lhs: SmartBits, rhs: _SmartExpr) -> (SmartBits, _SmartExpr):
    rhs = copy.deepcopy(rhs)
    rhs.resolve(Context(lhs, rhs))
    res = rhs.eval()
    return SmartBits.from_bits(res), rhs


def eval(expr: _SmartExpr, width: int, signed: bool = False):
    lhs = SmartBits[width, signed]()
    lhs @= expr
    return lhs


class _SmartifyTypeTransformer(TypeTransformer):
    def visit_Bits(self, T):
        signed = issint(T)
        return SmartBits[len(T), signed]


class _LeafCollector(ValueVisitor):
    def __init__(self):
        self.leaves = []

    def visit_Digital(self, value):
        self.leaves.append(value)

    def visit_Bits(self, value):
        self.leaves.append(value)


def make_smart(value):
    T = type(value)
    Tsmart = _SmartifyTypeTransformer().visit(T)
    Tsmart = Tsmart.qualify(Direction.Undirected)
    smart_value = Tsmart()
    leaf_collector = _LeafCollector()
    leaf_collector.visit(value)
    for leaf in leaf_collector.leaves:
        selector = make_selector(leaf)
        smart_leaf = selector.select(smart_value)
        smart_leaf @= leaf
    return smart_value



from abc import ABC, abstractmethod
from ..is_definition import isdefinition
from .tsort import tsort

__all__ = ['Pass', 'InstancePass', 'DefinitionPass', 'InstanceGraphPass',
           'pass_lambda']


class Pass(ABC):
    """Abstract base class for all passes"""
    def __init__(self, main):
        self.main = main

    def run(self):
        self.done()
        return self

    def done(self):
        pass


class InstancePass(Pass):
    def __init__(self, main):
        super().__init__(main)
        self.instances = []

    def _run(self, defn, path):
        for inst in defn.instances:
            inst_defn = type(inst)
            inst_path = path + (inst, )
            self.instances.append(inst_path)
            if callable(self):
                self(inst_path)
            if isdefinition(inst_defn):
                self._run(inst_defn, inst_path)

    def run(self):
        self._run(self.main, tuple())
        self.done()
        return self


class CircuitPass(Pass):
    """
    Run on all circuits (not just definitions)
    """
    def __init__(self, main):
        super().__init__(main)
        self.circuits = {}

    def _run(self, circuit):
        if isdefinition(circuit):
            for inst in circuit.instances:
                self._run(type(inst))
        # Call each definition only once.
        id_ = id(circuit)
        if id_ not in self.circuits:
            self.circuits[id_] = circuit
            if callable(self):
                self(circuit)

    def run(self):
        self._run(self.main)
        self.done()
        return self


class DefinitionPass(CircuitPass):
    """
    Run only only on circuits with definitions
    """
    def _run(self, circuit):
        if not isdefinition(circuit):
            return
        super()._run(circuit)


class BuildInstanceGraphPass(DefinitionPass):
    def __init__(self, main):
        super().__init__(main)
        self.graph = {}

    def __call__(self, defn):
        if defn not in self.graph:
            self.graph[defn] = []
        for inst in defn.instances:
            inst_defn = type(inst)
            if inst_defn not in self.graph:
                self.graph[inst_defn] = []
            if inst_defn not in self.graph[defn]:
                self.graph[defn].append(inst_defn)

    def done(self):
        graph = []
        for vert, edges in self.graph.items():
            graph.append((vert, edges))
        self.tsortedgraph = tsort(graph)


class InstanceGraphPass(Pass):
    def __init__(self, main):
        super(InstanceGraphPass, self).__init__(main)

        pass_ = BuildInstanceGraphPass(main).run()
        self.tsortedgraph = pass_.tsortedgraph

        if callable(self):
            for vert, edges in self.tsortedgraph:
                self(vert, edges)


class EditDefinitionPass(DefinitionPass):
    @abstractmethod
    def edit(self, circuit):
        raise NotImplementedError()

    def __call__(self, circuit):
        with circuit.open():
            self.edit(circuit)


# Auxiliary method to convert a simple pass, into a functional form, such as:
#
#   SomePass(ckt).run() <-> some_pass(ckt)
#
def pass_lambda(cls):

    def _fn(main, *args, **kwargs):
        p = cls(main, *args, **kwargs)
        p.run()
        return p

    return _fn



from magma.config import config, EnvConfig
from magma.backend.coreir.coreir_runtime import coreir_context, module_map
from magma.backend.coreir.coreir_transformer import DefnOrDeclTransformer
from magma.backend.coreir.insert_wrap_casts import insert_wrap_casts
from magma.logging import root_logger


config._register(
    coreir_backend_log_level=EnvConfig(
        "MAGMA_COREIR_BACKEND_LOG_LEVEL", "WARN"),
)

_logger = root_logger().getChild("coreir_backend")
_logger.setLevel(config.coreir_backend_log_level)


class CoreIRBackend:
    def __init__(self, context=None):
        self._init(context)

    def _init(self, context):
        singleton = coreir_context()
        if context is None:
            context = singleton
        if context is not singleton:
            _logger.warning("Creating CoreIRBackend with non-singleton CoreIR "
                            "context.")
        self._namespaces = module_map().setdefault(context, {})
        self._context = context
        self._lib_cache = {}
        self._included_libs = set()
        self._bound_modules = {}

    def add_module(self, magma_module, coreir_module):
        modules = self._namespaces.setdefault(magma_module.coreir_lib, {})
        modules[magma_module.coreir_name] = coreir_module

    def get_module(self, magma_module):
        # Throws KeyError if @magma_module or its namespace (coreir_lib) has
        # not been added.
        modules = self._namespaces[magma_module.coreir_lib]
        return modules[magma_module.coreir_name]

    def include_lib_or_libs(self, lib_or_libs):
        try:
            self._included_libs |= lib_or_libs
        except TypeError:  # single element
            self._included_libs.add(lib_or_libs)

    def included_libs(self):
        return self._included_libs.copy()

    def get_lib(self, lib):
        try:
            return self._lib_cache[lib]
        except KeyError:
            ret = self._lib_cache[lib] = self.context.get_lib(lib)
            return ret

    def bind_module(self, name, module):
        self._bound_modules[name] = module

    def bound_modules(self):
        return self._bound_modules.copy()

    @property
    def context(self):
        return self._context

    def reset(self):
        self._init(context=None)

    def compile(self, defn_or_decl, opts=None):
        _logger.debug(f"Compiling: {defn_or_decl.name}")
        opts = opts if opts is not None else {}
        transformer = DefnOrDeclTransformer(self, opts, defn_or_decl)
        transformer.run()
        return self._namespaces


def compile(main, file_name=None, context=None):
    insert_wrap_casts(main)
    backend = CoreIRBackend(context)
    backend.compile(main)
    if file_name is None:
        return backend.get_module(main)
    return backend.get_module(main).save_to_file(file_name)
